"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h
from multiprocessing import current_process as CurrentProcess

import json_any as json
from mpss_tools_36.constant.generic import REQUEST_CLOSE, REQUEST_STATUS, REQUEST_STOP
from mpss_tools_36.constant.process import MAIN_PROCESS_NAME
from mpss_tools_36.constant.wrapper import (
    REQUEST_ATTRIBUTE,
    REQUEST_ATTRIBUTES,
    REQUEST_CALL,
)
from mpss_tools_36.hint.wrapper import server_status_h
from mpss_tools_36.type.client.wrapper import client_t
from mpss_tools_36.type.line import line_t
from mpss_tools_36.type.server.generic import server_t as base_t


@d.dataclass(slots=True, repr=False, eq=False)
class server_t(base_t):
    object: h.Any = None
    # None default added because mpss_tools_36.type.server.generic.server_t has
    #     attributes with defaults.

    @property
    def status(self) -> server_status_h:
        """"""
        return self.name, type(self.object).__name__, self._thread is not None

    def __post_init__(self) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME

        reserved = client_t.ReservedAttributes()
        in_object = set(dir(self.object))
        assert in_object.intersection(reserved).__len__() == 0, (
            sorted(in_object),
            reserved,
            sorted(in_object.intersection(reserved)),
        )

    def _Run(self) -> None:
        """"""
        active_lines = list(self._lines)  # Work on a copy.
        while active_lines.__len__() > 0:
            for request_line_end, answer_line_end, request in base_t.ActiveRequests(
                active_lines
            ):
                if isinstance(request, tuple):
                    request, attribute, *args = request
                    if (args.__len__() > 0) and isinstance(args[-1], dict):
                        args, kwargs = args[:-1], args[-1]
                    else:
                        kwargs = {}
                else:
                    attribute, args, kwargs = None, (), {}
                # assert isinstance(request, str)

                output, still_to_be_sent = None, True
                if request == REQUEST_CALL:
                    output = getattr(self.object, attribute)(*args, **kwargs)
                elif request == REQUEST_ATTRIBUTE:
                    output = getattr(self.object, attribute)
                elif request == REQUEST_ATTRIBUTES:
                    output = tuple(
                        _
                        for _ in dir(self.object)
                        if not isinstance(getattr(self.object, _), h.Callable)
                    )
                elif request == REQUEST_STATUS:
                    answer_line_end.put(self.status)
                    still_to_be_sent = False
                elif request == REQUEST_CLOSE:
                    active_lines.remove(
                        line_t(request=request_line_end, answer=answer_line_end)
                    )
                    still_to_be_sent = False
                elif request == REQUEST_STOP:
                    active_lines.clear()
                    break
                else:
                    raise ValueError(
                        f'Unknown request "{request}" sent to '
                        f'data server "{self.name}".'
                    )

                if still_to_be_sent:
                    jsoned, issues = json.JsonStringOf(output)
                    if issues is None:
                        answer_line_end.put(jsoned)
                    else:
                        raise RuntimeError(
                            f"Un-jsonable object:\n{type(output).__name__}\n"
                            + "\n".join(issues)
                        )


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
from multiprocessing import Queue as NewQueue
from multiprocessing import current_process as CurrentProcess
from queue import Empty as EmptyError
from threading import Thread as thread_t

from mpss_tools_36.constant.generic import REQUEST_STOP
from mpss_tools_36.constant.process import MAIN_PROCESS_NAME
from mpss_tools_36.extension.data_class import NO_CHOICE
from mpss_tools_36.type.line import line_end_t, line_t


@d.dataclass(slots=True, repr=False, eq=False)
class server_t:
    name: str

    _lines: list[line_t] = NO_CHOICE(list)
    _thread: thread_t | None = NO_CHOICE(None)

    def NewLine(self) -> tuple[line_end_t, line_end_t]:
        """"""
        assert self._thread is None

        request_line_end, answer_line_end = NewQueue(), NewQueue()
        self._lines.append(line_t(request=request_line_end, answer=answer_line_end))

        return request_line_end, answer_line_end

    def Start(self) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME
        assert self._thread is None
        assert self._lines.__len__() > 0

        self._thread = thread_t(target=self._Run)
        self._thread.start()

    def _Run(self) -> None:
        """"""
        raise NotImplementedError

    @staticmethod
    def ActiveRequests(
        active_lines: list[line_t], /
    ) -> list[tuple[line_end_t, line_end_t, str | tuple]]:
        """"""
        output = []

        for line in tuple(active_lines):
            request_line_end = line.request
            try:
                request = request_line_end.get(False)
            except EmptyError:
                pass
            else:
                output.append((request_line_end, line.answer, request))

        return output

    def Stop(self, *_, **__) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME
        assert self._thread is not None

        # There are necessarily lines (see Start).
        self._lines[0].request.put(REQUEST_STOP)
        self._thread.join()
        self._thread = None

        for line_ends in self._lines:
            for line_end in line_ends:
                line_end.close()
                line_end.join_thread()
        self._lines.clear()


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import multiprocessing as prll
import time
import typing as h
from multiprocessing import Process as process_t
from multiprocessing import Value as shared_value_t
from multiprocessing import current_process as CurrentProcess
from threading import Thread as thread_t

from logger_36 import L
from mpss_tools_36.constant.feedback import DONE, TO_DO, UNKNOWN_VALUE
from mpss_tools_36.constant.process import MAIN_PROCESS_NAME
from mpss_tools_36.extension.chronos import FormatedDuration
from mpss_tools_36.extension.data_class import NO_CHOICE
from mpss_tools_36.hint.feedback import send_feedback_h


@d.dataclass(slots=True, repr=False, eq=False)
class server_t:
    n_iterations_per_task: int | h.Sequence[int]

    prefix: str = ""
    feedback_period: float = 3.0
    print_report: bool = False

    _message_template: str = d.field(init=False)
    _shared_iterations_per_task: list[h.Any] = NO_CHOICE(list)
    _shared_main_counter_per_task: list[h.Any] = NO_CHOICE(list)
    _thread: thread_t | None = NO_CHOICE(None)

    def __post_init__(self) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME

        self._message_template = (
            self.prefix + "|{}{: >6.2f}% {}| @ {}s/it #{} +{} -{}:{}"
        )

    def NewFeedbackSendingFunction(self) -> send_feedback_h | None:
        """"""
        if prll.get_start_method() != "fork":
            # forkserver and spawn need to pickle _SendFeedback, which they can since it
            # is a local function.
            return None

        shared_iterations, shared_main_counter = (
            shared_value_t("Q"),
            shared_value_t("Q"),
        )

        self._shared_iterations_per_task.append(shared_iterations)
        self._shared_main_counter_per_task.append(shared_main_counter)

        def _SendFeedback(iterations: int, main_counter: int, /) -> None:
            #
            shared_iterations.value = iterations
            shared_main_counter.value = main_counter

        return _SendFeedback

    def Start(self) -> None:
        """
        For sequential processing.
        """
        process = CurrentProcess()
        assert process.name == MAIN_PROCESS_NAME
        assert self._thread is None

        L.ToggleLogHolding(True)

        self._thread = thread_t(target=self.RunUntilExhausted, args=((process,),))
        self._thread.start()

    def RunUntilExhausted(self, tasks: h.Sequence[process_t], /) -> None:
        """
        For parallel processing (when called directly).
        """
        reference = start_time = time.time()

        n_tasks = self._shared_iterations_per_task.__len__()
        if n_tasks == 0:  # See NewFeedbackSendingFunction.
            return

        assert n_tasks == tasks.__len__()

        if isinstance(self.n_iterations_per_task, int):
            self.n_iterations_per_task = n_tasks * (self.n_iterations_per_task,)
        else:
            assert self.n_iterations_per_task.__len__() == n_tasks, (
                self.n_iterations_per_task.__len__(),
                n_tasks,
            )
        n_iterations_total = float(sum(self.n_iterations_per_task))
        if n_iterations_total == 0.0:
            return

        # See below about format.
        half_bar = 20 * TO_DO
        message = self._message_template.format(
            half_bar,
            0.0,
            half_bar,
            UNKNOWN_VALUE,
            UNKNOWN_VALUE,
            "00",
            UNKNOWN_VALUE,
            n_tasks,
        )
        message_length = message.__len__()
        print(f"{message}\r", end="", flush=True)

        while True:
            n_iterations_completed, total_main_counter = 0.0, 0
            n_active_tasks = n_tasks
            for task, n_iterations, shared_iterations, shared_main_counter in zip(
                tasks,
                self.n_iterations_per_task,
                self._shared_iterations_per_task,
                self._shared_main_counter_per_task,
                strict=True,
            ):
                iterations = shared_iterations.value
                if (iterations < 0) or (
                    (iterations < n_iterations) and not task.is_alive()
                ):
                    iterations = n_iterations
                    shared_iterations.value = n_iterations
                n_iterations_completed += iterations
                total_main_counter += shared_main_counter.value

                if iterations == n_iterations:
                    n_active_tasks -= 1

            if n_active_tasks == 0:
                break

            now = time.time()
            if now - reference < self.feedback_period:
                time.sleep(1.0)
                continue
            reference = now

            completion = n_iterations_completed / n_iterations_total

            elapsed_time = now - start_time
            if completion > 0.0:
                remaining_time = (elapsed_time / completion) - elapsed_time
                remaining_time = FormatedDuration(remaining_time)
            else:
                remaining_time = UNKNOWN_VALUE
            if n_iterations_completed > 0.0:
                period = n_tasks * elapsed_time / n_iterations_completed
                period = f"{period:.2f}"
            else:
                period = UNKNOWN_VALUE

            # Below, "X"=DONE and "-"=TO_DO.
            # 48 = 20_X_or_- (as in [:20])
            #    + 6_percent (as in 6.2f) + 1_% + 1_space
            #    + 20_X_or_- (as in [-20:]).
            n_completed = int(round(48.0 * completion))
            bar = f"{n_completed * DONE}{(48 - n_completed) * TO_DO}"
            message = self._message_template.format(
                bar[:20],
                100.0 * completion,
                bar[-20:],
                period,
                total_main_counter,
                FormatedDuration(elapsed_time),
                remaining_time,
                n_active_tasks,
            )
            message_length = max(message_length, message.__len__())
            print(f"{message: <{message_length}}\r", end="", flush=True)

            time.sleep(1.0)

        if self.print_report:
            elapsed_time = time.time() - start_time
            period = n_tasks * elapsed_time / n_iterations_total
            message = (
                f"{self.prefix}Summary: "
                f"{period:.2f}s/it "
                f"#{total_main_counter} "
                f"+{FormatedDuration(elapsed_time)}"
            )
            print(f"{message: <{message_length}}", flush=True)
        else:
            space = " "
            print(f"{space: <{message_length}}\r", end="", flush=True)

    def Stop(self) -> None:
        """
        For sequential processing.
        Safe to call on an un-started server.
        """
        assert CurrentProcess().name == MAIN_PROCESS_NAME

        if self._thread is not None:
            self._thread.join()
            self._thread = None

        L.ToggleLogHolding(False)

    def ReportTaskStarting(self, tasks: int | h.Sequence[process_t], /) -> None:
        """"""
        if isinstance(tasks, int):
            print(f"{self.prefix}Starting {tasks} tasks...", flush=True)
        else:
            pids = tuple(_.pid for _ in tasks)
            print(f"{self.prefix}All tasks started: {str(pids)[1:-1]}.", flush=True)


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h

import json_any as json
from mpss_tools_36.constant.data import (
    DELETION_DONE,
    REQUEST_CHECK,
    REQUEST_DELETE,
    REQUEST_GET,
    REQUEST_SET,
    REQUEST_SIZE,
    STORAGE_DONE,
)
from mpss_tools_36.constant.generic import REQUEST_STATUS
from mpss_tools_36.hint.data import server_status_h
from mpss_tools_36.type.line import line_end_t
from mpss_tools_36.type.server.data import server_t


@d.dataclass(slots=True, repr=False, eq=False)
class client_t:
    @classmethod
    def New(
        cls,
        *,
        server: server_t | None = None,
        request_line_end: line_end_t | None = None,
        answer_line_end: line_end_t | None = None,
    ) -> h.Self:
        """"""
        if server is None:
            return client_remote_t(
                request_line_end=request_line_end, answer_line_end=answer_line_end
            )
        return client_direct_t(server=server)


@d.dataclass(slots=True, repr=False, eq=False)
class client_direct_t(client_t):
    server: server_t

    def RequestedData(self, *names) -> h.Any | tuple[h.Any, ...]:
        """"""
        return self.server.RequestedData(*names)

    def Send(self, /, **data) -> None:
        """"""
        self.server.Store(**data)

    def RequestDeletion(self, *names) -> None:
        """"""
        self.server.Delete(*names)

    def ServerHasData(self, *names) -> bool | tuple[bool, ...]:
        """"""
        return self.server.Has(*names)

    def SizeOfServer(self) -> int:
        """"""
        return self.server.size

    def StatusOfServer(self) -> server_status_h:
        """"""
        return self.server.status


@d.dataclass(slots=True, repr=False, eq=False)
class client_remote_t(client_t):
    request_line_end: line_end_t
    answer_line_end: line_end_t

    def RequestedData(self, *names) -> h.Any | tuple[h.Any, ...]:
        """"""
        if names.__len__() == 0:
            request = REQUEST_GET
        else:
            request = (REQUEST_GET, *names)
        self.request_line_end.put(request)

        jsoned = self.answer_line_end.get()
        un_jsoned, issues = json.ObjectFromJsonString(jsoned)

        if issues is None:
            return un_jsoned

        raise RuntimeError(f"Un-de-jsonable object:\n{jsoned}\n" + "\n".join(issues))

    def Send(self, /, **data) -> None:
        """"""
        jsoned, issues = json.JsonStringOf(data)
        if issues is None:
            self.request_line_end.put((REQUEST_SET, jsoned))
        else:
            raise RuntimeError(f"Un-jsonable object:\n{data}\n" + "\n".join(issues))

        # Wait for setting confirmation for synchronization purposes. Thus, it will not
        # be possible to set again the same element before the current storage is
        # effective.
        acknowledgment = self.answer_line_end.get()
        if acknowledgment != STORAGE_DONE:
            raise RuntimeError(
                f"Unexpected storage acknowledgment received: {acknowledgment}"
            )

    def RequestDeletion(self, *names) -> None:
        """"""
        self.request_line_end.put((REQUEST_DELETE, *names))

        # Wait for deletion confirmation for synchronization purposes. Thus, the server
        # will not confirm the presence of in-the-course-of-deletion data.
        acknowledgment = self.answer_line_end.get()
        if acknowledgment != DELETION_DONE:
            raise RuntimeError(
                f"Unexpected deletion acknowledgment received: {acknowledgment}"
            )

    def ServerHasData(self, *names) -> bool | tuple[bool, ...]:
        """"""
        if names.__len__() == 0:
            request = REQUEST_CHECK
        else:
            request = (REQUEST_CHECK, *names)
        self.request_line_end.put(request)

        return self.answer_line_end.get()

    def SizeOfServer(self) -> int:
        """"""
        self.request_line_end.put(REQUEST_SIZE)
        return self.answer_line_end.get()

    def StatusOfServer(self) -> server_status_h:
        """"""
        self.request_line_end.put(REQUEST_STATUS)
        return self.answer_line_end.get()


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

from multiprocessing import process
from multiprocessing.managers import BaseProxy as base_t
from multiprocessing.managers import Token, dispatch, listener_client, public_methods
from traceback import format_exc


def all_attributes(obj):
    """"""
    temp = []
    for name in dir(obj):
        func = getattr(obj, name)
        if not callable(func):
            temp.append(name)
    return temp


def public_attributes(obj):
    """"""
    return ["." + name for name in all_attributes(obj) if name[0] != "_"]


def serve_client(self, conn):
    """"""
    recv = conn.recv
    send = conn.send
    id_to_obj = self.id_to_obj

    while not self.stop_event.is_set():
        try:
            methodname = obj = None
            request = recv()
            ident, methodname, args, kwds = request
            try:
                obj, exposed, gettypeid = id_to_obj[ident]
            except KeyError as ke:
                try:
                    obj, exposed, gettypeid = self.id_to_local_proxy_obj[ident]
                except KeyError:
                    raise ke

            if is_attribute := (methodname[0] == "."):
                methodname = methodname[1:]
            function = getattr(obj, methodname)

            try:
                if is_attribute:
                    res = function
                else:
                    res = function(*args, **kwds)
            except Exception as e:
                msg = ("#ERROR", e)
            else:
                typeid = gettypeid and gettypeid.get(methodname, None)
                if typeid:
                    rident, rexposed = self.create(conn, typeid, res)
                    token = Token(typeid, self.address, rident)
                    msg = ("#PROXY", (rexposed, token))
                else:
                    msg = ("#RETURN", res)

        except AttributeError:
            if methodname is None:
                msg = ("#TRACEBACK", format_exc())
            else:
                try:
                    fallback_func = self.fallback_mapping[methodname]
                    result = fallback_func(self, conn, ident, obj, *args, **kwds)
                    msg = ("#RETURN", result)
                except Exception:
                    msg = ("#TRACEBACK", format_exc())

        except EOFError:
            import sys

            sys.exit(0)

        except Exception:
            msg = ("#TRACEBACK", format_exc())

        try:
            try:
                send(msg)
            except Exception:
                send(("#UNSERIALIZABLE", format_exc()))
        except Exception as e:
            import sys

            conn.close()
            sys.exit(1)


def create(self, c, typeid, /, *args, **kwds):
    """"""
    with self.mutex:
        callable, exposed, method_to_typeid, proxytype = self.registry[typeid]

        if callable is None:
            if kwds or (len(args) != 1):
                raise ValueError("Without callable, must have one non-keyword argument")
            obj = args[0]
        else:
            obj = callable(*args, **kwds)

        if exposed is None:
            exposed = public_methods(obj) + public_attributes(obj)
        if method_to_typeid is not None:
            if not isinstance(method_to_typeid, dict):
                raise TypeError(
                    "Method_to_typeid {0!r}: type {1!s}, not dict".format(
                        method_to_typeid, type(method_to_typeid)
                    )
                )
            exposed = list(exposed) + list(method_to_typeid)

        ident = "%x" % id(obj)  # convert to string because xmlrpclib
        # only has 32 bit signed integers

        self.id_to_obj[ident] = (obj, set(exposed), method_to_typeid)
        if ident not in self.id_to_refcount:
            self.id_to_refcount[ident] = 0

    self.incref(c, ident)
    return ident, tuple(exposed)


def MakeProxyType(name, exposed, _cache={}):
    """"""
    exposed = tuple(exposed)
    try:
        return _cache[(name, exposed)]
    except KeyError:
        pass

    dic = {}

    for meth in exposed:
        if meth[0] == ".":
            exec(
                """@property
def %s(self):
            return self._callmethod(%r, (), {})"""
                % (meth[1:], meth),
                dic,
            )
        else:
            exec(
                """def %s(self, /, *args, **kwds):
            return self._callmethod(%r, args, kwds)"""
                % (meth, meth),
                dic,
            )

    ProxyType = type(name, (base_t,), dic)
    ProxyType._exposed_ = exposed
    _cache[(name, exposed)] = ProxyType
    return ProxyType


def ProxyWithAttributes(
    token,
    serializer,
    manager=None,
    authkey=None,
    exposed=None,
    incref=True,
    manager_owned=False,
):
    """"""
    _Client = listener_client[serializer][1]

    if exposed is None:
        conn = _Client(token.address, authkey=authkey)
        try:
            exposed = dispatch(conn, None, "get_methods", (token,))
        finally:
            conn.close()

    if authkey is None and manager is not None:
        authkey = manager._authkey
    if authkey is None:
        authkey = process.current_process().authkey

    ProxyType = MakeProxyType("ProxyWithAttributes[%s]" % token.typeid, exposed)
    proxy = ProxyType(
        token,
        serializer,
        manager=manager,
        authkey=authkey,
        incref=incref,
        manager_owned=manager_owned,
    )
    proxy._isauto = True
    return proxy


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import multiprocessing as prll
import os as o
import tempfile as tmpf
import time
import types as t
import typing as h
from multiprocessing.managers import BaseManager as base_t
from pathlib import Path as path_t
from threading import Thread as thread_t

from mpss_tools_36.constant.process import MAIN_PROCESS_NAME
from mpss_tools_36.extension.data_class import NO_CHOICE
from mpss_tools_36.extension.multiprocessing_manager import (
    ProxyWithAttributes,
    create,
    serve_client,
)

LOCAL_HOST = "127.0.0.1"
PROXY_CONNEXION_PORT = "PROXY_CONNEXION_PORT"
PROXY_CONNEXION_EXTENSION = ".connexion"
SHARING_PROXY = "SharingProxy"


@d.dataclass(init=False, slots=True, repr=False, eq=False)
class sharer_t(base_t):
    _instances: h.ClassVar[dict[str, h.Any]] = {}

    thread: thread_t | None = NO_CHOICE(None)

    @classmethod
    def _AddInstance(cls, instance: h.Any, name: str, /) -> None:
        """"""
        assert prll.current_process().name == MAIN_PROCESS_NAME
        assert name not in cls._instances
        cls._instances[name] = instance

    @classmethod
    def _GetInstance(cls, name: str, /) -> h.Any:
        """"""
        return cls._instances[name]

    def __init__(self, *args, **kwargs) -> None:
        """"""
        cls = self.__class__
        cls.register(
            SHARING_PROXY, callable=cls._GetInstance, proxytype=ProxyWithAttributes
        )
        base_t.__init__(self, *args, **kwargs)
        self.thread = None  # Because default __init__ is not used.

    @classmethod
    def New(cls, /, *, instance: h.Any = None, name: str | None = None) -> h.Self:
        """"""
        assert prll.current_process().name == MAIN_PROCESS_NAME

        if instance is not None:
            assert name is not None
            cls._AddInstance(instance, name)

        return cls(address=(LOCAL_HOST, 0))

    def Add(self, instance: h.Any, name: str, /) -> None:
        """"""
        assert prll.current_process().name == MAIN_PROCESS_NAME
        self.__class__._AddInstance(instance, name)

    def Start(self) -> None:
        """"""
        assert prll.current_process().name == MAIN_PROCESS_NAME

        server = self.get_server()
        server.serve_client = t.MethodType(serve_client, server)
        server.create = t.MethodType(create, server)

        self.thread = thread_t(target=server.serve_forever, daemon=True)
        self.thread.start()

        port = str(server.address[1])
        if prll.get_start_method() != "spawn":
            o.environ[PROXY_CONNEXION_PORT] = port
        else:
            # With the "spawn" start method, changes to os.environ are not inherited.
            with tmpf.NamedTemporaryFile(
                prefix=f"{PROXY_CONNEXION_PORT}-{time.time_ns()}-",
                suffix=PROXY_CONNEXION_EXTENSION,
                delete=False,
                mode="w",
            ) as accessor:
                accessor.write(port)

    @classmethod
    def Address(cls) -> tuple[str, int]:
        """"""
        if prll.get_start_method() != "spawn":
            return LOCAL_HOST, int(o.environ[PROXY_CONNEXION_PORT])

        return LOCAL_HOST, int(cls.PathOfPort().read_text())

    @staticmethod
    def PathOfPort() -> path_t:
        """"""
        paths = sorted(
            path_t(tmpf.gettempdir()).glob(
                f"{PROXY_CONNEXION_PORT}-*-*{PROXY_CONNEXION_EXTENSION}"
            )
        )
        assert paths.__len__() > 0

        return paths[-1]

    @classmethod
    def SharedInstanceProxy(cls, name: str, /) -> h.Any:
        """"""
        sharer = cls(address=cls.Address())
        sharer.connect()

        return getattr(sharer, SHARING_PROXY)(name)

    def Stop(self) -> None:
        """"""
        assert prll.current_process().name == MAIN_PROCESS_NAME

        cls = self.__class__

        cls._instances.clear()
        self.thread = None

        if prll.get_start_method() != "spawn":
            # del o.environ[PROXY_CONNEXION_PORT]  # Raises KeyError(?)
            o.environ[PROXY_CONNEXION_PORT] = ""
        else:
            cls.PathOfPort().unlink()


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""



"""
Copyright CNRS (https://www.cnrs.fr/index.php/en)
Contributor(s): Eric Debreuve (eric.debreuve@cnrs.fr) since 2025
SEE COPYRIGHT NOTICE BELOW
"""

import dataclasses as d
import typing as h
from multiprocessing import current_process as CurrentProcess
from multiprocessing.shared_memory import SharedMemory as shared_memory_t

import json_any as json
import numpy as nmpy
from logger_36.api.memory import ObjectSize
from mpss_tools_36.constant.data import (
    DELETION_DONE,
    REQUEST_CHECK,
    REQUEST_DELETE,
    REQUEST_GET,
    REQUEST_SET,
    REQUEST_SIZE,
    STORAGE_DONE,
)
from mpss_tools_36.constant.generic import REQUEST_CLOSE, REQUEST_STATUS, REQUEST_STOP
from mpss_tools_36.constant.process import MAIN_PROCESS_NAME
from mpss_tools_36.extension.data_class import NO_CHOICE
from mpss_tools_36.hint.data import server_status_h
from mpss_tools_36.task.numpy_sharing import DisposeOriginalSharedArray, NewSharedArray
from mpss_tools_36.type.line import line_t
from mpss_tools_36.type.server.generic import server_t as base_t

array_t = nmpy.ndarray


@d.dataclass(slots=True, repr=False, eq=False)
class server_t(base_t):
    """
    Numpy arrays must be set (using REQUEST_SET_NUMPY) and requested alone.

    The type of data can be h.Any at instantiation time, in which case it is considered
    as an object which will be converted into an attribute-name/attribute-value
    dictionary.
    """

    data: dict[str, h.Any] | h.Any = d.field(default_factory=dict)
    is_read_only: bool = True

    _shared_array_memory: dict[str, shared_memory_t] = NO_CHOICE(dict)

    @property
    def size(self) -> int:
        """"""
        return ObjectSize(self) + self.shared_size

    @property
    def shared_size(self) -> int:
        """"""
        return sum(_.size for _ in self._shared_array_memory.values())

    @property
    def status(self) -> server_status_h:
        """"""
        shared_array_memory = {
            _.name: f"SHARED:{_.size:_}" for _ in self._shared_array_memory.values()
        }

        return (
            self.name,
            self.size,
            {
                _: f"{type(__).__module__}.{type(__).__name__}"
                for _, __ in self.data.items()
            }
            | shared_array_memory,
            self.is_read_only,
            self._thread is not None,
        )

    def __post_init__(self) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME

        if isinstance(self.data, dict):
            return

        data = {}
        for attribute in dir(self.data):
            if attribute[0] != "_":
                data[attribute] = getattr(self.data, attribute)
        self.data = data

    def _Run(self) -> None:
        """"""
        active_lines = list(self._lines)  # Work on a copy.
        while active_lines.__len__() > 0:
            for request_line_end, answer_line_end, request in base_t.ActiveRequests(
                active_lines
            ):
                if isinstance(request, tuple):
                    request, *arguments = request
                else:
                    arguments = ()
                # assert isinstance(request, str)

                if request == REQUEST_GET:
                    # arguments must be a typing.Sequence of str.
                    data = self.RequestedData(*arguments)
                    jsoned, issues = json.JsonStringOf(data)
                    if issues is None:
                        answer_line_end.put(jsoned)
                    else:
                        raise RuntimeError(
                            f"Un-jsonable object:\n{type(data).__name__}\n"
                            + "\n".join(issues)
                        )
                elif request == REQUEST_SET:
                    # arguments must be a typing.Sequence with a single, jsoned
                    # dict[str, typing.Any].
                    un_jsoned, issues = json.ObjectFromJsonString(arguments[0])
                    if issues is None:
                        self.Store(**un_jsoned)
                        answer_line_end.put(STORAGE_DONE)  # For synchronization.
                    else:
                        raise RuntimeError(
                            f"Un-de-jsonable object:\n{arguments[0]}\n"
                            + "\n".join(issues)
                        )
                elif request == REQUEST_DELETE:
                    # arguments must be a typing.Sequence of str.
                    self.Delete(*arguments)
                    answer_line_end.put(DELETION_DONE)  # For synchronization.
                elif request == REQUEST_CHECK:
                    # arguments must be a typing.Sequence of str.
                    answer_line_end.put(self.Has(*arguments))
                elif request == REQUEST_SIZE:
                    answer_line_end.put(self.size)
                elif request == REQUEST_STATUS:
                    answer_line_end.put(self.status)
                elif request == REQUEST_CLOSE:
                    active_lines.remove(
                        line_t(request=request_line_end, answer=answer_line_end)
                    )
                elif request == REQUEST_STOP:
                    active_lines.clear()
                    break
                else:
                    raise ValueError(
                        f'Unknown request "{request}" sent to '
                        f'data server "{self.name}".'
                    )

    def Has(self, *names) -> bool | tuple[bool, ...]:
        """"""
        if (n_names := names.__len__()) == 0:
            return self.data.__len__() > 0

        if n_names == 1:
            return names[0] in self.data

        return tuple(_ in self.data for _ in names)

    def RequestedData(self, *names) -> h.Any | tuple[h.Any, ...]:
        """
        If called indirectly (as a result of a request from a subprocess), returned data
        must be jsoned.
        """
        if (n_names := names.__len__()) == 0:
            return self.data

        if n_names == 1:
            return self.data[names[0]]

        return tuple(self.data[_] for _ in names)

    def Store(self, /, **data) -> None:
        """
        If called indirectly (as a result of a request from a subprocess), data in
        arguments have been jsoned.
        """
        if self.is_read_only:
            raise RuntimeError(
                f'Attempt to modify read-only data server "{self.name}".'
            )

        for name, value in data.items():
            if isinstance(value, array_t):
                if name in self._shared_array_memory:
                    DisposeOriginalSharedArray(self._shared_array_memory[name])
                _, sharing_name, shared_memory = NewSharedArray(value)
                self.data[name] = sharing_name
                self._shared_array_memory[name] = shared_memory
            else:
                self.data[name] = value

    def Delete(self, *names) -> None:
        """"""
        if names.__len__() == 0:
            names = tuple(self.data.keys())

        for name in names:
            if name in self._shared_array_memory:
                DisposeOriginalSharedArray(self._shared_array_memory[name])
                del self._shared_array_memory[name]
            del self.data[name]

    def DisposeSharedResources(self) -> None:
        """"""
        assert CurrentProcess().name == MAIN_PROCESS_NAME

        for shared_memory in self._shared_array_memory.values():
            DisposeOriginalSharedArray(shared_memory)
        self._shared_array_memory.clear()

    def Stop(self, /, *, should_dispose_shared_resources: bool = True) -> None:
        """"""
        base_t.Stop(self)

        if should_dispose_shared_resources:
            self.DisposeSharedResources()


"""
COPYRIGHT NOTICE

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability.

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or
data to be ensured and,  more generally, to use and operate it in the
same conditions as regards security.

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

SEE LICENCE NOTICE: file README-LICENCE-utf8.txt at project source root.
"""
